//! Contains files to create and build queries to send latter to lektord.

use crate::*;
use amadeus_macro::*;
use std::string::ToString;

pub(crate) enum LektorQueryLineType {
    Ok,
    ListOk,
    Ack,
    Continuation(usize),
    Data,
}

#[derive(Debug, Clone)]
pub enum LektorQuery {
    Ping,
    Close,
    KillServer,
    ConnectAsUser(String, Box<LektorQuery>),

    CurrentKara,
    PlaybackStatus,

    PlayNext,
    PlayPrevious,
    ShuffleQueue,

    ListAllPlaylists,
    ListPlaylist(String),
    SearchKara(LektorUri),

    FindAddKara(LektorUri),
    InsertKara(LektorUri),
    AddKara(LektorUri),

    Continuation(usize, Box<LektorQuery>),
}

impl std::str::FromStr for LektorQueryLineType {
    type Err = ();

    fn from_str(line: &str) -> Result<Self, Self::Err> {
        if Self::is_line_ok(line) {
            Ok(Self::Ok)
        } else if Self::is_line_ack(line) {
            Ok(Self::Ack)
        } else if Self::is_line_list_ok(line) {
            Ok(Self::ListOk)
        } else if let Some(cont) = Self::is_line_continuation(line) {
            Ok(Self::Continuation(cont))
        } else {
            Ok(Self::Data)
        }
    }
}

impl LektorQueryLineType {
    fn is_line_continuation(line: &str) -> Option<usize> {
        if !line.starts_with("continue:") {
            return None;
        }
        match line.trim_start_matches("continue:").trim().parse::<usize>() {
            Ok(cont) => Some(cont),
            Err(_) => None,
        }
    }

    fn is_line_list_ok(line: &str) -> bool {
        (line == "list_OK\n") || (line == "list_OK")
    }

    fn is_line_ok(line: &str) -> bool {
        (line == "OK\n") || (line == "OK")
    }

    fn is_line_ack(line: &str) -> bool {
        line.starts_with("ACK: ")
    }
}

/// The type of the function to use to produce a typed response from the
/// formated response.
type QueryToTypeResponseBuilder = fn(LektorFormatedResponse) -> Result<LektorResponse, String>;

impl LektorQuery {
    /// Get the function to use to produce the typed response from the formated
    /// response, to automatically create the correct thing.
    pub fn get_response_type(&self) -> QueryToTypeResponseBuilder {
        use LektorQuery::*;
        match self {
            Ping | Close | KillServer | PlayNext | PlayPrevious | ShuffleQueue | InsertKara(_)
            | AddKara(_) => LektorEmptyResponse::from_formated,

            ListAllPlaylists => LektorPlaylistSetResponse::from_formated,
            PlaybackStatus => LektorPlaybackStatusResponse::from_formated,
            CurrentKara => LektorCurrentKaraResponse::from_formated,

            ConnectAsUser(_, cmd) | Continuation(_, cmd) => cmd.get_response_type(),

            ListPlaylist(_) | SearchKara(_) | FindAddKara(_) => {
                LektorKaraSetResponse::from_formated
            }
        }
    }

    /// Create a continued query out of another one. If the query is already a
    /// continuation query then the underlying query is reused.
    pub fn create_continuation(query: Self, cont: usize) -> Self {
        match query {
            Self::Continuation(_, query) => Self::Continuation(cont, query),
            _ => Self::Continuation(cont, Box::new(query)),
        }
    }

    /// Verify that a query is Ok.
    pub fn verify(&self) -> Result<(), String> {
        use LektorQuery::*;
        match self {
            // User commands
            SearchKara(_) | FindAddKara(_) | InsertKara(_) | AddKara(_) | PlaybackStatus
            | PlayNext | PlayPrevious | ShuffleQueue | ListAllPlaylists | ListPlaylist(_)
            | CurrentKara | Ping => Ok(()),

            // Should be admin commands
            Close => Err("close is an admin command".to_string()),
            KillServer => Err("kill server is an admin command".to_string()),

            // Admin commands
            ConnectAsUser(_, cmd) => match cmd.as_ref() {
                Close | KillServer => Ok(()),
                _ => Err(format!("not an admin command: {cmd:?}")),
            },

            // Continuation commands
            Continuation(_, cmd) => match cmd.as_ref() {
                ListAllPlaylists | FindAddKara(_) | SearchKara(_) | ListPlaylist(_) => Ok(()),
                _ => Err(format!("not a continuable command: {cmd:?}")),
            },
        }
    }
}

impl ToString for LektorQuery {
    fn to_string(&self) -> String {
        use LektorQuery::*;
        match self {
            Ping => lkt_command_from_str!("ping"),
            Close => lkt_command_from_str!("close"),
            KillServer => lkt_command_from_str!("kill"),
            ConnectAsUser(password, cmd) => format!(
                concat!(
                    "command_list_ok_begin\n",
                    "password {}\n",
                    "{}\n",
                    "command_list_end\n",
                ),
                password,
                cmd.to_string()
            ),

            CurrentKara => lkt_command_from_str!("currentsong"),
            PlaybackStatus => lkt_command_from_str!("status"),

            PlayNext => lkt_command_from_str!("next"),
            PlayPrevious => lkt_command_from_str!("previous"),
            ShuffleQueue => lkt_command_from_str!("shuffle"),

            ListAllPlaylists => lkt_command_from_str!("listplaylists"),
            ListPlaylist(plt_name) => format!("listplaylist {}\n", plt_name),
            SearchKara(uri) => format!("find {}\n", uri.to_string()),

            FindAddKara(uri) => format!("findadd {}\n", uri.to_string()),
            InsertKara(uri) => format!("__insert {}\n", uri.to_string()),
            AddKara(uri) => format!("add {}\n", uri.to_string()),

            Continuation(cont, query) => format!("{cont} {}", query.to_owned().to_string()),
        }
    }
}
